/**
 * @file HuffmanCoder.h
 * Declaration of the HuffmanCoder class.
 * Version: 23.02.2009
 * Author: Ruediger Knoerig
 */

#ifndef HUFFMANCODER_H_
#define HUFFMANCODER_H_

#include <itpp/base/vec.h>
#include <itpp/base/array.h>

#ifdef DEBUG
#include <iostream>

using namespace std;

#endif

namespace itpp {

/**
 * @brief Huffman Coder implementation class.
 * This class provides a Huffman entropy coder.
 * @author Ruediger Knoerig
 */
class HuffmanCoder {
public:

	/**
	 * @brief Default constructor resulting in an uninitialized and untrained Huffman coder.
	 */
	HuffmanCoder() : initialized(false),trained(false) {}

	/**
	 * @brief Default constructor resulting in an initialized but untrained Huffman coder.
	 * @param H0 The entropy coder will have an input range of \f$[0\ldots 2^{H0}-1\f$.
	 */
	HuffmanCoder(unsigned int H0) { setRawRate(H0); }

	/**
	 * @brief Default constructor resulting in an initialized and untrained Huffman coder.
	 * @param H0 The entropy coder will have an input range of \f$[0\ldots 2^{H0}-1\f$.
	 */
	HuffmanCoder(unsigned int H0,const itpp::Vec<unsigned int> &trainingdata) { setRawRate(H0); train(trainingdata); }

	/**
	 * @brief Destructor.
	 */
	virtual ~HuffmanCoder() {}

	/**
	 * @brief Setter for the H0 rate / raw PCM rate.
	 * After calling this method the Huffman coder will be in untrained state again!
	 * @param H0 The entropy coder will have an input range of \f$[0\ldots 2^{H0}-1\f$.
	 */

	inline void setRawRate(unsigned int H0);
	/**
	 * @brief Method for coding an individual input sample.
	 * @param symbol Symbol code in raw PCM.
	 * @return Bit vector containing the Huffman code according to the input sample.
	 */

	inline itpp::bvec code(unsigned int symbol);

	/**
	 * @brief Method for coding an input vector.
	 * @param symbols Raw PCM input vector.
	 * @return Bit vector containing the Huffman-coded input vector.
	 */
	inline itpp::bvec code(itpp::Vec<unsigned int> symbols);


	/**
	 * @brief Method for decoding an coded vector
	 * @param coded Huffman-Coded bit sequence.
	 * @return decoded Vector filled with the decoded sequence.
	 */
	itpp::Vec<unsigned int> decode(itpp::bvec coded);

	/**
	 * @brief Method for training the Huffman coder.
	 * @param trainingdata Vector of input data to train the Huffman coder.
	 * @return mean code length.
	 */
	double train(const Vec<unsigned int> &trainingdata);

	/**
	 * @brief Getter for the trained state.
	 * @return True if the Huffman coder has been trained.
	 */
	bool isTrained() const { return trained; }

	/**
	 * @brief Getter for the initialized state.
	 * @return True if the Huffman coder has been initialized.
	 */
	bool isInitialized() const { return initialized; }

	/**
	 * @brief Getter for the upper input range delimiter (input range: \f$[0\ldots N-1]\f$).
	 * @return N
	 */
	unsigned int getN() const { return N; }

	/**
	 * @brief Getter for the raw PCM input rate the Huffman coder is build for.
	 * @return H0 / raw PCM rate.
	 */
	unsigned int getH0() const { return H0; }

	/**
	 * @brief Getter for the coding table.
	 * The coding table consists of the Huffman code vector for each input symbol.
	 * @return The N Huffman code vectors (array index = according symbol).
	 */
	const itpp::Array<itpp::bvec> &getCodeTable() const { return codetable; }

	/**
	 * @brief Getter for the decoding table.
	 * Address the vector with the numbers of '0' in the huffman code and you get
	 * the according symbol.
	 */
	const itpp::Vec<unsigned int> &getDecodeTable() const { return decodetable; }

protected:
	bool initialized; /**< Flag indicating if the Huffman coder has been initialized with a raw rate */
	bool trained; /**< Flag indicating if the Huffman coder has been trained */
	unsigned int N; /**< input range [0..N-1] */
	unsigned int H0; /**< Raw PCM rate */
	itpp::Array<itpp::bvec> codetable; /**< encoding table */
	itpp::Vec<unsigned int> decodetable; /**< decoding table */

private:
	/**
	 * @brief Internal class to create the huffman table.
	 */
	class sort_elem{
	public:
		/**
		 * @brief Constructor.
		 */
		sort_elem() { rel_h = 0; }

		/**
		 * @brief destructor
		 */
		~sort_elem() {}

		/**
		 * @brief Setter for the symbol value.
		 * @param symbol Symbol value for the element.
		 */
		void setSymbol(unsigned int symbol) { this->symbol=symbol; }

		/**
		 * @brief getter for the symbol value.
		 * @return Symbol value of the element.
		 */
		unsigned int getSymbol() const { return symbol; }

		/**
		 * @brief getter for the frequency of the symbol value.
		 * @return Frequency of occurence of the element.
		 */
		unsigned int getFrequency() const { return rel_h; }

		/**
		 * @brief Increment operator for the frequency.
		 */
		void operator++() { rel_h++; }

		/**
		 * @brief Compare operator <
		 */
		friend bool operator<(const HuffmanCoder::sort_elem &a,const HuffmanCoder::sort_elem &b) { return (a.rel_h < b.rel_h); }

		/**
		 * @brief Compare operator <=
		 */
		friend bool operator<=(const HuffmanCoder::sort_elem &a,const HuffmanCoder::sort_elem &b) { return (a.rel_h <= b.rel_h); }

		/**
		 * @brief Compare operator >
		 */
		friend bool operator>(const HuffmanCoder::sort_elem &a,const HuffmanCoder::sort_elem &b) { return (a.rel_h > b.rel_h); }

		/**
		 * @brief Compare operator >=
		 */
		friend bool operator>=(const HuffmanCoder::sort_elem &a,const HuffmanCoder::sort_elem &b) { return (a.rel_h >= b.rel_h); }

		/**
		 * @brief Compare operator ==
		 */
		friend bool operator==(const HuffmanCoder::sort_elem &a,const HuffmanCoder::sort_elem &b) { return (a.rel_h == b.rel_h); }

		/**
		 * @brief Compare operator !=
		 */
		friend bool operator!=(const HuffmanCoder::sort_elem &a,const HuffmanCoder::sort_elem &b) { return (a.rel_h != b.rel_h); }

	protected:
		unsigned int symbol; /**< symbol value of the element */
		unsigned int rel_h; /**< frequency of the symbol */
	};
};

itpp::bvec HuffmanCoder::code(unsigned int symbol)
{
#ifdef DEBUG
		if(!initialized) cerr << "HuffmanCoder::code() called on an uninitialized object!" << endl;
		if(!initialized) cerr << "HuffmanCoder::code() called on an untrained object!" << endl;
		if(symbol >= N) cerr << "HuffmanCoder::code(symbol): symbol value " << symbol << "out of range [0.." << N-1 << "]!" << endl;
#endif
return codetable(symbol);
}

itpp::bvec HuffmanCoder::code(itpp::Vec<unsigned int> symbols)
	{
		itpp::bvec output;
		for(int k=0;k<symbols.length();k++) output=concat(output,code(symbols(k)));
		return output;
	}
}

void itpp::HuffmanCoder::setRawRate(unsigned int H0)
{
	this->H0=H0;
	N=1<<H0;
	codetable.set_size(N,false);
	decodetable.set_size(N,false);
	trained=false;
	initialized=true;
}

#endif /* HUFFMANCODER_H_ */

/** \example main.cpp
 * Example code for using this class.
 */

/*! \mainpage Hauptseite
 *
 * \section intro_sec Introduction
 * The HuffmanCoder class implements an adaptive Huffman Coder/Decoder.
 */
